#!/usr/bin/python3
"""
Copyright (C) 2013, 2014 Johan Mattsson

This library is free software; you can redistribute it and/or modify 
it under the terms of the GNU Lesser General Public License as 
published by the Free Software Foundation; either version 3 of the 
License, or (at your option) any later version.

This library is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of 
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
Lesser General Public License for more details.
"""

VERSION = '2.21.1'
SO_VERSION_MAJOR = '36'
SO_VERSION_MINOR = '0'
SO_VERSION = SO_VERSION_MAJOR + '.' + SO_VERSION_MINOR

LIBBIRDGEMS_SO_VERSION_MAJOR = '0'
LIBBIRDGEMS_SO_VERSION_MINOR = '0'
LIBBIRDGEMS_SO_VERSION = LIBBIRDGEMS_SO_VERSION_MAJOR + '.' + LIBBIRDGEMS_SO_VERSION_MINOR;
